import os
from datetime import datetime, timezone
import numpy as np
import logging
import logging.handlers

app_name = 'sonarcal'

def setupLogging(log_dir, label):
    """Set info, warning, and error message logger to a file and to the console."""
    now = datetime.now(timezone.utc)
    logger_filename = os.path.join(log_dir, now.strftime('log_' + label + '-%Y%m%d-T%H%M%S.log'))
    logger = logging.getLogger(app_name)
    logger.setLevel(logging.INFO)

    formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')

    # A logger to a file that is changed periodically
    rotatingFile = logging.handlers.TimedRotatingFileHandler(logger_filename, when='H',
                                                             interval=12, utc=True)
    rotatingFile.setFormatter(formatter)
    logger.addHandler(rotatingFile)

    # add a second output to the logger to direct messages to the console
    console = logging.StreamHandler()
    console.setFormatter(formatter)
    logger.addHandler(console)

    logger.info('Log files are in %s.', log_dir.as_posix())


def on_exit(root, job, sig):
    """Call when the Windows cmd console closes."""
    root.after_cancel(job)
    logging.info('Program ending...')
    root.quit()
    # not sure why this call is needed...
    window_closed(root, job)


def window_closed(root, job):
    """Call to nicely end the whole program."""
    root.after_cancel(job)
    logging.info('Program ending...')
    logging.shutdown()  # not working???
    root.quit()


def beamAnglesFromNetCDF4(f, beamGroup, i):
    """Calculate the beam angles as per the convention for the given beamGroup and ping index."""
    x = f[beamGroup + '/beam_direction_x'][i]
    y = f[beamGroup + '/beam_direction_y'][i]
    z = f[beamGroup + '/beam_direction_z'][i]
    tilt = np.arctan(z / np.sqrt(x**2 + y**2))  # [rad]

    # convert x,y,z direction into a horizontal angle for use elsewhere
    theta = np.arctan2(-y, x)
    # Make angles go 0 to 2pi, not -pi to 0 to pi (all anti-clockwise)
    theta = np.mod(theta, 2*np.pi)

    return theta, tilt


def SvFromSonarNetCDF4(f, beamGroup, i, tilt):
    """Calculate Sv from the given beam group and ping."""
    eqn_type = f[beamGroup].attrs['conversion_equation_type']
    # work around the current Simrad files using integers instead of the
    # type defined in the convetion (which shows up here as a string)
    if isinstance(eqn_type, np.ndarray):
        eqn_type = f'type_{eqn_type[0]}'
    else:
        eqn_type = eqn_type.decode('utf-8')

    if eqn_type == 'type_2':

        # Pick out various variables for the given ping, i
        sv = f[beamGroup + '/backscatter_r'][i]  # an array for each beam
        tau_e = f[beamGroup + '/transmit_duration_equivalent'][i]  # a scaler for the current ping
        Psi = f[beamGroup + '/equivalent_beam_angle'][i]  # a scalar for each beam
        SL = f[beamGroup + '/transmit_source_level'][i]  # a scalar for the current ping
        K = f[beamGroup + '/receiver_sensitivity'][i]  # a scalar for each beam
        deltaG = f[beamGroup + '/gain_correction'][i]  # a scalar for each beam
        G_T = f[beamGroup + '/time_varied_gain'][i]  # a value for each sample in the current ping
        ping_freq_1 = f[beamGroup + '/transmit_frequency_start'][i]  # a scalar for each beam
        ping_freq_2 = f[beamGroup + '/transmit_frequency_stop'][i]  # a scalar for each beam

        # and some more constant things that could be moved out of this function...
        c = f['Environment/sound_speed_indicative'][()]  # a scalar
        alpha_vector = f['Environment/absorption_indicative'][()]  # a vector
        freq_vector = f['Environment/frequency'][()]  # a vector
        ping_freq = (ping_freq_1 + ping_freq_2)/2.0  # a scalar for each beam
        alpha = np.interp(ping_freq, freq_vector, alpha_vector)  # a scalar for each beam

        # some files have nan for some of the above variables, so fix that
        if np.any(np.isnan(deltaG)):
            deltaG = np.zeros(deltaG.shape)
        if np.any(np.isnan(alpha_vector)):
            # quick and dirty...
            alpha = acousticAbsorption(10.0, 35.0, 10.0, ping_freq)

        a = 10.0 * np.log10(c * tau_e * Psi / 2.0) + SL + K + deltaG  # a scalar for each beam
        r_offset = 0.25 * c * tau_e

        samInt = f[beamGroup + '/sample_interval'][i]  # [s]

        # usually some zeros in the data of no real consequence
        with np.errstate(divide='ignore', invalid='ignore'):
            for j in range(0, sv.shape[0]):  # loop over each beam
                # [m] range vector for the current beam
                r = samInt * c/2.0 * np.arange(0, sv[j].size) - r_offset
                sv[j] = 20.0*np.log10(sv[j]/np.sqrt(2.0)) + 20.0*np.log10(r)\
                    + 2*alpha[j]*r - a[j] + G_T

    elif eqn_type == 'type_1':
        # Pick out various variables for the given ping, i
        p_r = f[beamGroup + '/backscatter_r'][i]  # an array for each beam
        p_i = f[beamGroup + '/backscatter_i'][i]  # an array for each beam
        sv = np.absolute(p_r + 1j*p_i)
        tau_e = f[beamGroup + '/transmit_duration_equivalent'][i]  # a scaler for the current ping
        Psi = f[beamGroup + '/equivalent_beam_angle'][i]  # a scalar for each beam
        G = f[beamGroup + '/transducer_gain'][i]  # a scalar for each beam
        P = f[beamGroup + '/transmit_power'][i]  # a scalar
        ping_freq_1 = f[beamGroup + '/transmit_frequency_start'][i]  # a scalar for each beam
        ping_freq_2 = f[beamGroup + '/transmit_frequency_stop'][i]  # a scalar for each beam

        # and some more constant things that could be moved out of this function...
        c = f['Environment/sound_speed_indicative'][()]  # a scalar
        alpha_vector = f['Environment/absorption_indicative'][()]  # a vector
        freq_vector = f['Environment/frequency'][()]  # a vector
        ping_freq = (ping_freq_1 + ping_freq_2)/2.0  # a scalar for each beam
        alpha = np.interp(ping_freq, freq_vector, alpha_vector)  # a scalar
        wl = c / ping_freq  # wavelength [m]

        if np.any(np.isnan(alpha_vector)):
            # quick and dirty...
            alpha = acousticAbsorption(10.0, 35.0, 10.0, ping_freq)

        samInt = f[beamGroup + '/sample_interval'][i]  # [s]

        r_offset = 0.0  # incase we need this in the future

        # usually some zeros in the data of no real consequence
        with np.errstate(divide='ignore', invalid='ignore'):
            for k in range(0, sv.shape[0]):  # loop over each beam
                # [m] range vector for the current beam
                r = samInt * c/2.0 * np.arange(0, sv[k].size) - r_offset
                sv[k] = 20.0*np.log10(sv[k]) + 20.0*np.log10(r) + 2*alpha*r\
                    - 10.0*np.log10((P*wl*wl*c*Psi[k]*tau_e) / (32*np.pi*np.pi))\
                    - G[k] - 40.0*np.log10(np.cos(tilt[k]))

    else:  # unsupported format - just take the log10 of the numbers. Usually usefull.
        sv = f[beamGroup + '/backscatter_r'][i]
        with np.errstate(divide='ignore'):
            for j in range(0, sv.shape[0]):
                sv[j] = np.log10(sv[j])

    return sv


def acousticAbsorption(temperature, salinity, depth, frequency):
    """Calculate acoustic absorption.

    Uses Ainslie & McColm, 1998.
    Units are:
        temperature - degC
        salinity - PSU
        depth - m
        frequency - Hz
        alpha - dB/m
    """
    frequency = frequency / 1e3  # [kHz]
    pH = 8.0

    z = depth/1e3  # [km]
    f1 = 0.78 * np.sqrt(salinity/35.0) * np.exp(temperature/26.0)
    f2 = 42.0 * np.exp(temperature/17.0)
    alpha = 0.106 * (f1*frequency**2./(frequency**2+f1**2)) * np.exp((pH-8.0)/0.56) \
        + 0.52*(1+temperature/43.0) * (salinity/35.0) \
        * (f2*frequency**2)/(frequency**2+f2**2) * np.exp(z/6.0) \
        + 0.00049*frequency**2 * np.exp(-(temperature/27.0+z/17.0))
    alpha = alpha * 1e-3  # [dB/m]

    return alpha



import sys
from time import sleep
from datetime import datetime, timedelta
import h5py
from .utils import beamAnglesFromNetCDF4, SvFromSonarNetCDF4, app_name
import logging


logger = logging.getLogger(app_name)

def most_recent_file(watch_dir, wait_interval):
    """Get the most recent .nc file in the directory."""

    while True:
        files = sorted(list(watch_dir.glob('*.nc')))
        if files:
            return files[-1]

        logger.info("No .nc file found in '%s'", watch_dir)
        sleep(wait_interval)

def file_type(watch_dir):
    """Works out what sonar the data came from, and what format it is."""
    
    # Options will be 'sonar-netcdf4', 'CS90-raw', 'SN90-raw'

    return 'sonar-netcdf4'

def file_listen(watchDir, beamGroup, msg_queue):
    """ """
    f_type = file_type(watchDir)
    
    params = (watchDir, beamGroup, msg_queue)
    
    match f_type:
        case 'sonar-netcdf4':
            file_listen_netcdf(*params)
        case 'CS90-raw':
            file_listen_cs90_raw(*params)
        case 'SN90-raw':
            file_listen_sn90_raw(*params)
        case _:
            logger.error('Unsupported file type')


def file_replay(watchDir, beamGroup, msg_queue, replayRate):
    """ """
    f_type = file_type(watchDir)

    params = (watchDir, beamGroup, msg_queue, replayRate)
    
    match f_type:
        case 'sonar-netcdf4':
            file_replay_netcdf(*params)
        case 'CS90-raw':
            file_replay_cs90_raw(*params)
        case 'SN90-raw':
            file_replay_sn90_raw(*params)
        case _:
            logger.error('Unsupported file type')



def file_listen_netcdf(watchDir, beamGroup, msg_queue):
    """Listen for new data in a file.

    Find new data in the most recent file (and keep checking for more new data).
    Used for live calibrations.
    """
    # A more elegant method for all of this can be found in the examples here:
    # https://docs.h5py.org/en/stable/swmr.html, which uses the watch facility
    # in the hdf5 library (but we're not sure if the omnisonars write data in
    # a manner that this will work with).

    # Config how and when to give up looking for new data in an existing file.
    maxNoNewDataCount = 20  # number of tries to find new pings in an existing file
    waitInterval = 0.5  # [s] time period between checking for new pings
    waitIntervalFile = 1.0  # [s] time period between checking for new files
    errorWaitInterval = 0.2  # [s] time period to wait if there is a file read error

    pingIndex = -1  # which ping to read. -1 means the last ping, -2 the second to last ping

    t_previous = 0  # timestamp of previous ping
    f_previous = ''  # previously used file

    while True:  # could add a timeout on this loop...
        mostRecentFile = most_recent_file(watchDir, waitIntervalFile)

        if mostRecentFile == f_previous:  # no new file was found
            logger.info('No newer file found. Will try again in %s s.', str(waitIntervalFile))
            sleep(waitIntervalFile)  # wait and try again
        else:
            logger.info('Listening to file: %s.', mostRecentFile)
            noNewDataCount = 0

            while noNewDataCount <= maxNoNewDataCount:
                # open netcdf file
                try:
                    f = h5py.File(mostRecentFile, 'r', libver='latest', swmr=True)
                    # f = h5py.File(mostRecentFile, 'r') # without HDF5 swmr option
                    f_previous = mostRecentFile

                    t = f[beamGroup + '/ping_time'][pingIndex]

                    if t > t_previous:  # there is a new ping in the file
                        pingTime = datetime(1601, 1, 1) + timedelta(microseconds=t/1000.0)
                        logger.info('Start reading ping from time %s', pingTime)

                        theta, tilt = beamAnglesFromNetCDF4(f, beamGroup, pingIndex)
                        sv = SvFromSonarNetCDF4(f, beamGroup, pingIndex, tilt)

                        samInt = f[beamGroup + '/sample_interval'][pingIndex]
                        c = f['Environment/sound_speed_indicative'][()]
                        labels = f[beamGroup + '/beam']

                        t_previous = t
                        noNewDataCount = 0  # reset the count

                        logger.info('Finished reading ping from time %s', pingTime)
                        # send the data off to be plotted
                        msg_queue.put((t, samInt, c, sv, theta, labels))
                    else:
                        noNewDataCount += 1
                        if noNewDataCount > maxNoNewDataCount:
                            logger.info('No new data found in file %s after waiting %.1f s.',
                                         mostRecentFile.name, noNewDataCount * waitInterval)

                    f.close()
                    # try this instead of opening and closing the file
                    # t.id.refresh(), etc
                    sleep(waitInterval)
                except OSError:
                    f.close()  # just in case...
                    e = sys.exc_info()
                    logger.warning('OSError when reading netCDF4 file:')
                    logger.warning(e)
                    logger.warning('Ignoring the above and trying again.')
                    sleep(errorWaitInterval)


def file_replay_netcdf(watchDir, beamGroup, msg_queue, replayRate):
    """Replay all data in the newest file. Used for testing."""
    waitIntervalFile = 1.0  # [s] time period between checking for new files

    mostRecentFile = most_recent_file(watchDir, waitIntervalFile)

    logger.info('Reading from file: %s.', mostRecentFile)

    # open netcdf file
    f = h5py.File(mostRecentFile, 'r')

    t = f[beamGroup + '/ping_time']

    # Send off each ping at a sedate rate...
    for i in range(0, t.shape[0]):
        # print('ping')
        theta, tilt = beamAnglesFromNetCDF4(f, beamGroup, i)
        sv = SvFromSonarNetCDF4(f, beamGroup, i, tilt)

        samInt = f[beamGroup + '/sample_interval'][i]
        c = f['Environment/sound_speed_indicative'][()]
        labels = f[beamGroup + '/beam']

        # send the data off to be plotted
        msg_queue.put((t[i], samInt, c, sv, theta, labels))

        # Ping at recorded ping rate if asked
        if replayRate == 'realtime' and i > 0:
            # t has units of nanoseconds
            sleep((t[i] - t[i-1])/1e9)
        else:
            sleep(0.2)

    f.close()

    logger.info('Finished replaying file: %s', mostRecentFile)


def file_replay_cs90_raw(watchDir, beamGroup, msg_queue, replayRate):
    """Replay all data in the newest file. Used for testing."""

    logger.error('CS90 raw files are not yet supported')


def file_replay_sn90_raw(watchDir, beamGroup, msg_queue, replayRate):
    """Replay all data in the newest file. Used for testing."""

    logger.error('SN90 raw files are not yet supported')


def file_listen_cs90_raw(watchDir, beamGroup, msg_queue):
    """XXX"""
    logger.error('CS90 raw files are not yet supported')
    

def file_listen_sn90_raw(watchDir, beamGroup, msg_queue):
    """XXX"""
    logger.error('SN90 raw files are not yet supported')



import traceback
from queue import Empty
from datetime import datetime, timedelta, timezone
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
from scipy import signal
from matplotlib.widgets import RangeSlider
from .utils import app_name
from .gui_utils import draggable_ring, draggable_radial
import humanize
import logging

# Matplotlib for tkinter
mpl.use('TkAgg')

logger = logging.getLogger(app_name)

class echogramPlotter:
    """Receive via a queue new ping data and use that to update the display."""

    def __init__(self, numPings, maxRange, maxSv, minSv, msg_queue, root):
        self.queue = msg_queue
        self.root = root
        self.job = None

        # All callback that is called whenever a new ping has finished drawing
        self.new_ping_cb = None

        # Various user-changable lines on the plots that could in the future
        # come from a config file.
        self.beamLineAngle = 0.0  # [deg]
        self.beam = 0  # dummy value. Is updated once some data are received.

        self.minTargetRange = 0.33*maxRange
        self.maxTargetRange = 0.66*maxRange

        self.varNum = 5  # number of sphere values to use for the ping-to-ping variability

        self.diffPlotXlim = (-3, 0)  # [dB]

        self.numPings = numPings  # to show in the echograms
        self.maxRange = maxRange  # [m] of the echograms
        self.maxSv = maxSv  # [dB] max Sv to show in the echograms
        self.minSv = minSv  # [dB] min Sv to show in the echograms

        self.checkQueueInterval = 200  # [ms] duration between checking the queue for new data

        self.movingAveragePoints = 10  # number of points for moving average for smoothed plots

        self.emptySv = -999.0  # initialisation value of echogram data

        # Make the plots. It gets filled with pretty things once the first ping
        # of data is received.
        self.fig = plt.figure(figsize=(11.5, 5))
        plt.ion()

        self.firstPing = True

    def createGUI(self, samInt, c, backscatter, theta, labels):
        """Create the GUI."""
        cmap = mpl.colormaps['jet']  # viridis looks nice too...
        cmap.set_under('w')  # and for values below self.minSv, if desired

        # the max extend of the threshold range slider
        lowestSv = -100
        highestSv = 10

        # number of samples to store per ping
        self.maxSamples = int(np.ceil(self.maxRange / (samInt*c/2.0)))
        self.numBeams = backscatter.shape[0]

        # Storage for the things we plot
        # Polar plot
        self.polar = np.ones((self.maxSamples, self.numBeams), dtype=float) * self.emptySv
        # Echograms
        self.port = np.ones((self.maxSamples, self.numPings), dtype=float) * self.emptySv
        self.main = np.ones((self.maxSamples, self.numPings), dtype=float) * self.emptySv
        self.stbd = np.ones((self.maxSamples, self.numPings), dtype=float) * self.emptySv
        # Amplitude of sphere
        self.amp = np.ones((3, self.numPings), dtype=float) * np.nan
        self.ampSmooth = np.ones((3, self.numPings), dtype=float) * np.nan
        # Range of the max amplitude within the range range on the selected beam
        self.rangeMax = None  # [m]

        # Differences in sphere amplitudes, smoothed version
        self.ampDiffPort = np.ones((self.numPings), dtype=float) * np.nan
        self.ampDiffStbd = np.ones((self.numPings), dtype=float) * np.nan
        self.ampDiffPortSmooth = np.ones((self.numPings), dtype=float) * np.nan
        self.ampDiffStbdSmooth = np.ones((self.numPings), dtype=float) * np.nan

        # Make the plot axes and set up static things
        self.polarPlotAx = plt.subplot2grid((3, 3), (0, 0), rowspan=3, projection='polar')
        self.portEchogramAx = plt.subplot2grid((3, 3), (0, 1))
        self.mainEchogramAx = plt.subplot2grid((3, 3), (1, 1))
        self.stbdEchogramAx = plt.subplot2grid((3, 3), (2, 1))
        self.ampPlotAx = plt.subplot2grid((3, 3), (0, 2), rowspan=2)
        self.ampDiffPlotAx = plt.subplot2grid((3, 3), (2, 2))

        plt.tight_layout(pad=4, w_pad=0.05, h_pad=0.05)

        # Configure the echogram axes
        self.portEchogramAx.invert_yaxis()
        self.mainEchogramAx.invert_yaxis()
        self.stbdEchogramAx.invert_yaxis()

        self.portEchogramAx.yaxis.tick_right()
        self.mainEchogramAx.yaxis.tick_right()
        self.stbdEchogramAx.yaxis.tick_right()

        self.portEchogramAx.xaxis.set_ticklabels([])
        self.mainEchogramAx.xaxis.set_ticklabels([])

        # Configure the sphere amplitude axes
        self.ampPlotAx.yaxis.tick_right()
        self.ampPlotAx.yaxis.set_label_position("right")
        self.ampPlotAx.xaxis.set_ticklabels([])
        self.ampPlotAx.grid(axis='y', linestyle=':')
        self.ampDiffPlotAx.yaxis.tick_right()
        self.ampDiffPlotAx.yaxis.set_label_position("right")
        self.ampDiffPlotAx.grid(axis='y', linestyle=':')

        self.portEchogramAx.set_title('Port', loc='left')
        self.mainEchogramAx.set_title(f'Beam {self.beam}', loc='left')
        self.stbdEchogramAx.set_title('Starboard', loc='left')

        # Create the lines in the plots
        # Sphere TS from 3 beams
        self.ampPlotLinePort, = self.ampPlotAx.plot(self.amp[0, :], 'r-', linewidth=1)
        self.ampPlotLineMain, = self.ampPlotAx.plot(self.amp[1, :], 'k-', linewidth=1)
        self.ampPlotLineStbd, = self.ampPlotAx.plot(self.amp[2, :], 'g-', linewidth=1)
     
        # Smoothed curves for the TS from 3 beams
        self.ampPlotLinePortSmooth, = self.ampPlotAx.plot(self.ampSmooth[0, :], 'r-', linewidth=2)
        self.ampPlotLineMainSmooth, = self.ampPlotAx.plot(self.ampSmooth[1, :], 'k-', linewidth=2)
        self.ampPlotLineStbdSmooth, = self.ampPlotAx.plot(self.ampSmooth[2, :], 'g-', linewidth=2)
        self.ampPlotAx.set_xlim(0, self.numPings)
     
        # a informative number on the TS plot
        self.diffVariability = self.ampPlotAx.text(0.05, 0.95, '', ha='left', va='top',
                                                   transform=self.ampPlotAx.transAxes)
        self.diffVariability.set_bbox({'color': 'w', 'alpha': 0.5})

        # Difference in sphere TS from the 3 beams
        self.ampDiffPortPlot, = self.ampDiffPlotAx.plot(self.ampDiffPort, 'r-', linewidth=1)
        self.ampDiffStbdPlot, = self.ampDiffPlotAx.plot(self.ampDiffStbd, 'g-', linewidth=1)
        # Smoothed curves of the difference in TS
        self.ampDiffPortPlotSmooth, = self.ampDiffPlotAx.plot(self.ampDiffPortSmooth, 'r-',
                                                              linewidth=2)
        self.ampDiffStbdPlotSmooth, = self.ampDiffPlotAx.plot(self.ampDiffStbdSmooth, 'g-',
                                                              linewidth=2)
        self.ampDiffPlotAx.set_xlim(0, self.numPings)
        self.ampDiffPlotAx.set_ylim(self.diffPlotXlim)

        # Echograms for the 3 selected beams
        ee = [0.0, self.numPings, self.maxRange, 0.0]
        self.portEchogram = self.portEchogramAx.imshow(self.port, interpolation='nearest',
                                                       aspect='auto', extent=ee, vmin=self.minSv,
                                                       vmax=self.maxSv)
        self.mainEchogram = self.mainEchogramAx.imshow(self.main, interpolation='nearest',
                                                       aspect='auto', extent=ee, vmin=self.minSv,
                                                       vmax=self.maxSv)
        self.stbdEchogram = self.stbdEchogramAx.imshow(self.stbd, interpolation='nearest',
                                                       aspect='auto', extent=ee, vmin=self.minSv,
                                                       vmax=self.maxSv)

        self.portEchogram.set_cmap(cmap)
        self.mainEchogram.set_cmap(cmap)
        self.stbdEchogram.set_cmap(cmap)

        # Omni echogram axes setup
        self.polarPlotAx.set_theta_offset(np.pi/2)  # to make bow direction plot upwards
        self.polarPlotAx.set_frame_on(False)
        self.polarPlotAx.xaxis.set_ticklabels([])

        # Omni echogram image
        r = np.arange(0, self.maxSamples)*samInt*c/2.0
        self.polarPlot = self.polarPlotAx.pcolormesh(theta, r, self.polar,
                                                     shading='auto', vmin=self.minSv,
                                                     vmax=self.maxSv)
        self.polarPlotAx.grid(axis='y', linestyle=':')

        self.polarPlot.set_cmap(cmap)

        # Colorbar for the omni echogram
        cb = plt.colorbar(self.polarPlot, ax=self.polarPlotAx, orientation='horizontal',
                          extend='both', fraction=0.05, location='bottom')
        cb.set_label('$S_v$ re 1 m$^{-1}$ [dB]')

        # range slider to adjust the echogram thresholds

        slider_ax = plt.axes([0.028, 0.20, 0.015, 0.65])
        self.slider = RangeSlider(slider_ax, "Thresholds", lowestSv, highestSv,
                                  valstep=np.arange(lowestSv, highestSv+1, 1),
                                  orientation='vertical', facecolor='blue')
        # using valinit in the constructor fails due to a bug, so use this workaround
        self.slider.set_val((self.minSv, self.maxSv))
        self.slider.on_changed(self.updateEchogramThresholds)

        # Range rings on the omni echogram
        self.rangeRing1 = draggable_ring(self.polarPlotAx, self.minTargetRange)
        self.rangeRing2 = draggable_ring(self.polarPlotAx, self.maxTargetRange)
        self.beamLine = draggable_radial(self.polarPlotAx, self.beamLineAngle,
                                         self.maxRange, theta, labels)

        self.updateBeamNum(theta)  # sets self.beam from the positon of the radial line

        # Axes labels
        self.stbdEchogramAx.set_xlabel('Pings')

        self.portEchogramAx.yaxis.set_label_position('right')

        self.mainEchogramAx.yaxis.set_label_position('right')
        self.mainEchogramAx.set_ylabel('Range (m)')

        self.stbdEchogramAx.yaxis.set_label_position('right')

        self.ampDiffPlotAx.set_xlabel('Pings')
        self.ampPlotAx.set_ylabel('$S_v$ re 1 m$^{-1}$ [dB]')
        self.ampDiffPlotAx.set_ylabel(r'$\Delta$ (dB)')
        self.ampPlotAx.set_title('Maximum amplitude at 0 m')

    def updateEchogramThresholds(self, val):
        """Update the image colormaps."""
        self.polarPlot.set_clim(val)
        self.portEchogram.set_clim(val)
        self.mainEchogram.set_clim(val)
        self.stbdEchogram.set_clim(val)

        # Redraw the figure to ensure it updates
        self.fig.canvas.draw_idle()

    def set_ping_callback(self, cb):
        """Set the callback that is called after each new ping is displayed."""
        self.new_ping_cb = cb

    def newPing(self, label):
        """Receive messages from the queue, decodes them and updates the echogram."""
        while not self.queue.empty():
            try:
                message = self.queue.get(block=False)
            except Empty:
                logger.info('No new data in received message.')
            else:
                try:
                    (t, samInt, c, backscatter, theta, labels) = message

                    if self.firstPing:
                        self.firstPing = False
                        self.createGUI(samInt, c, backscatter, theta, labels)

                    # Update the plots with the data in the new ping
                    pingTime = datetime(1601, 1, 1, tzinfo=timezone.utc)\
                        + timedelta(microseconds=t/1000.0)
                    timeBehind = datetime.now(timezone.utc) - pingTime
                    milliseconds = pingTime.microsecond / 1000
                    label.config(text=f'Ping at {pingTime:%Y-%m-%d %H:%M:%S}.' +
                                 f'{milliseconds:03.0f} '
                                 f'({humanize.precisedelta(timeBehind)} ago)')
                    logger.info('Displaying ping that occurred at %s.', pingTime)

                    self.minTargetRange = min(self.rangeRing1.range, self.rangeRing2.range)
                    self.maxTargetRange = max(self.rangeRing1.range, self.rangeRing2.range)

                    # print('Range rings: {}, {}'.format(self.minTargetRange, self.maxTargetRange))

                    minSample = int(np.floor(2*self.minTargetRange / (samInt * c)))
                    maxSample = int(np.floor(2*self.maxTargetRange / (samInt * c)))

                    self.updateBeamNum(theta)  # sets self.beam from self.beamLineAngle

                    # work out the beam indices
                    if self.beam == 0:
                        beamPort = self.numBeams-1
                    else:
                        beamPort = self.beam-1

                    if self.beam == self.numBeams-1:
                        beamStbd = 0
                    else:
                        beamStbd = self.beam+1

                    # print('{}, {}, {}'.format(beamPort, self.beam, beamStbd))
                    # Find the max amplitude between the min and max ranges set by the UI
                    # and store for plotting
                    self.amp = np.roll(self.amp, -1, 1)
                    self.amp[0, -1] = np.max(backscatter[beamPort][minSample:maxSample])
                    max_i = np.argmax(backscatter[self.beam][minSample:maxSample])
                    self.amp[1, -1] = backscatter[self.beam][minSample+max_i]
                    self.rangeMax = (minSample+max_i) * samInt * c / 2.0
                    self.amp[2, -1] = np.max(backscatter[beamStbd][minSample:maxSample])

                    # Store the amplitude for the 3 beams for the echograms
                    self.port = self.updateEchogramData(self.port, backscatter[beamPort])
                    self.main = self.updateEchogramData(self.main, backscatter[self.beam])
                    self.stbd = self.updateEchogramData(self.stbd, backscatter[beamStbd])

                    # Update the plots
                    # Sphere TS from 3 beams
                    self.ampPlotLinePort.set_ydata(self.amp[0, :])
                    self.ampPlotLineMain.set_ydata(self.amp[1, :])
                    self.ampPlotLineStbd.set_ydata(self.amp[2, :])
                    # and smoothed plots
                    coeff = np.ones(self.movingAveragePoints)/self.movingAveragePoints
                    # and measure of ping-to-ping variability
                    variability = np.std(self.amp[1, -self.varNum: -1])
                    if not np.isnan(variability):
                        self.diffVariability.set_text(rf'$\sigma$ = {variability:.1f} dB')

                    self.ampSmooth[0, :] = signal.filtfilt(coeff, 1, self.amp[0, :])
                    self.ampSmooth[1, :] = signal.filtfilt(coeff, 1, self.amp[1, :])
                    self.ampSmooth[2, :] = signal.filtfilt(coeff, 1, self.amp[2, :])
                    self.ampPlotLinePortSmooth.set_ydata(self.ampSmooth[0, :])
                    self.ampPlotLineMainSmooth.set_ydata(self.ampSmooth[1, :])
                    self.ampPlotLineStbdSmooth.set_ydata(self.ampSmooth[2, :])

                    self.ampPlotAx.set_title(f'Maximum amplitude at {self.rangeMax:.1f} m')
                    self.ampPlotAx.relim()
                    self.ampPlotAx.autoscale_view()

                    # Difference in sphere TS from 3 beams
                    diffPort = self.amp[0, :] - self.amp[1, :]
                    diffStbd = self.amp[2, :] - self.amp[1, :]
                    self.ampDiffPortPlot.set_ydata(diffPort)
                    self.ampDiffStbdPlot.set_ydata(diffStbd)
                    # and the smoothed
                    smPort = signal.filtfilt(coeff, 1, diffPort)
                    smStbd = signal.filtfilt(coeff, 1, diffStbd)
                    self.ampDiffPortPlotSmooth.set_ydata(smPort)
                    self.ampDiffStbdPlotSmooth.set_ydata(smStbd)

                    self.ampDiffPlotAx.relim()
                    self.ampDiffPlotAx.autoscale_view(scaley=False)

                    # Beam echograms
                    self.portEchogram.set_data(self.port)
                    self.mainEchogram.set_data(self.main)
                    self.stbdEchogram.set_data(self.stbd)

                    self.portEchogramAx.set_title(f'Beam {labels[beamPort].decode()}', loc='left')
                    self.mainEchogramAx.set_title(f'Beam {labels[self.beam].decode()}', loc='left')
                    self.stbdEchogramAx.set_title(f'Beam {labels[beamStbd].decode()}', loc='left')

                    # Polar plot
                    for i, b in enumerate(backscatter):
                        if b.shape[0] > self.maxSamples:
                            self.polar[:, i] = b[0: self.maxSamples]
                        else:
                            samples = b.shape[0]
                            self.polar[:, i] =\
                                np.concatenate((b, self.emptySv*np.ones(self.maxSamples-samples)),
                                               axis=0)

                    self.polarPlot.set_array(self.polar.ravel())

                    if self.new_ping_cb:
                        self.new_ping_cb()

                except Exception:  # if anything goes wrong, just ignore it...
                    logger.warning('Error when processing and displaying echo data:')
                    logger.warning(traceback.print_exc())
                    logger.warning('Ignoring the above and waiting for next ping.')

        self.job = self.root.after(self.checkQueueInterval, self.newPing, label)

    def updateEchogramData(self, data, pingData):
        """Shift the ping data to the left and add in the new ping data."""
        data = np.roll(data, -1, 1)
        if pingData.shape[0] > self.maxSamples:
            data[:, -1] = pingData[0:self.maxSamples]
        else:
            samples = pingData.shape[0]
            data[:, -1] = np.concatenate((pingData[:],
                                          self.emptySv*np.ones(self.maxSamples-samples)), axis=0)
        return data

    def updateBeamNum(self, theta):
        """Get the beam number from the beam line angle and the latest theta."""
        self.beamLineAngle = self.beamLine.value

        idx = (np.abs(theta - self.beamLineAngle)).argmin()
        self.beam = idx



"""Omnisonar calibration program

Provides omni and echogram displays and sphere amplitude plots for use when
calibrating omni-directional sonars.
"""
# TODO:
# Choose beam_group based on beam type rather than requiring it in the config file

import configparser
import tkinter as tk
from functools import partial
import threading
import queue
import sys
from pathlib import Path

from platformdirs import PlatformDirs

from .echogram_plotter import echogramPlotter
from .utils import setupLogging, app_name, on_exit, window_closed
from .file_ops import file_listen, file_replay
from .calibration_gui import calibrationGUI

if sys.platform == "win32":
    import win32api

# Configure logging
dirs = PlatformDirs(appname=app_name, appauthor="Aqualyd")
log_dir = Path(dirs.user_log_dir)
log_dir.mkdir(parents=True, exist_ok=True)
setupLogging(log_dir, app_name)


def main():
    """Omnisonar calibration graphical user interface."""    
    ##########################################
    # Sort out the configuration file
    config_filename = Path(dirs.user_config_dir)/'config.ini'
    config_filename.parent.mkdir(parents=True, exist_ok=True)
    
    config = configparser.ConfigParser()
    c = config.read(config_filename, encoding='utf8')

    if not c:  # config file not found, so make one
        config['DEFAULT'] = {'numPingsToShow': 100,
                             'maxRange': 50,
                             'maxSv': -20,
                             'minSv': -60,
                             'replayRate': 'realtime',
                             'horizontalBeamGroupPath': 'Sonar/Beam_group1',
                             'watchDir': '.',
                             'liveData': 'yes',
                             'helpURI': 'https://aqualyd-limited.github.io/sonarCal/'
                             }

        with open(config_filename, 'w', encoding='utf-8') as configfile:
            config.write(configfile)
        # TODO - open config dialog instead of exitting here
        print('No config file was found, so ' + str(config_filename) +
              ' was created. You may need to edit this file.')
        sys.exit()

    # Pull out the settings in the config file.
    numPings = config.getint('DEFAULT', 'numPingsToShow')
    maxRange = config.getfloat('DEFAULT', 'maxRange')
    maxSv = config.getfloat('DEFAULT', 'maxSv')
    minSv = config.getfloat('DEFAULT', 'minSv')
    replayRate = config.get('DEFAULT', 'replayRate')
    horizontalBeamGroup = config.get('DEFAULT', 'horizontalBeamGroupPath')
    watchDir = Path(config.get('DEFAULT', 'watchDir'))
    liveData = config.getboolean('DEFAULT', 'liveData')
    helpURI = config.get('DEFAULT', 'helpURI')

    ##########################################
    # Start things...

    # queue to communicate between two threads
    msg_queue = queue.Queue()
    
    # Tk GUI
    root = tk.Tk()

    # handle to the function that does the echogram drawing
    # job = None  

    echogram = echogramPlotter(numPings, maxRange, maxSv, minSv, msg_queue, root)
    gui = calibrationGUI(echogram, title='Sonar calibration', help_uri=helpURI)
    # Check periodically for new echogram data
    # job = root.after(echogram.checkQueueInterval, echogram.newPing, gui.status_label())

    # Start receive in a separate thread
    if liveData:
        t = threading.Thread(target=file_listen, args=(watchDir, horizontalBeamGroup, msg_queue))
    else:
        t = threading.Thread(target=file_replay, args=(watchDir, horizontalBeamGroup, msg_queue,
                                                       replayRate))
    t.daemon = True  # makes the thread close when main() ends
    t.start()

    # For Windows, catch when the console is closed
    if sys.platform == "win32":
        win32api.SetConsoleCtrlHandler(partial(on_exit, gui.root(), gui.job()), True)

    # And start things...
    root.protocol("WM_DELETE_WINDOW", lambda: window_closed(gui.root(), gui.job()))
    root.mainloop()





import pandas as pd

class calibrationData():
    """Storage for sonar caliration results."""

    def __init__(self):
        self.data = pd.DataFrame(columns=['Time', 'Gain (dB)', 'RMS (dB)', 'Range (m)', 'Echoes'])
        self.data.index.name = 'Beam'
    
    def update(self, beam: int, timestamp: str, gain: float, rms: float, r: float, num: int):
        self.data.loc[beam] = (timestamp, gain, rms, r, num)
        
    def remove(self, beams: list[int]):
        """Remove data for given beam."""
        self.data.drop(index=beams, inplace=True)
        
    def df(self):
        return self.data  # eventually return a better form of the data?
    


"""
Estimate calibration gains from sphere calibration data collected using Furuno omni-sonars.

@author: Gavin Macaulay, Institute of Marine Research, Norway
"""
# pylint: disable=invalid-name # too late to change all the variable names, etc.

from pathlib import Path
import configparser
import logging
import logging.handlers
import os
from datetime import datetime, timezone
import matplotlib.pyplot as plt
import scipy.stats.mstats as ms
import numpy as np
import h5py
import pandas as pd
import cftime

# The config file should be in the same directory as this script.
parent = Path(__file__).resolve().parent
configFilename = parent.joinpath('gain_calibration.ini')


def main():
    """Calculate the calibrated gain for each beam in the given beam file."""
    config = configparser.ConfigParser()
    config.read(configFilename, encoding='utf8')

    logDir = Path(config.get('DEFAULT', 'logDir'))
    calLogFile = Path(config.get('DEFAULT', 'calLogFile'))
    resultsDir = Path(config.get('DEFAULT', 'resultsDir'))
    offset = config.getfloat('DEFAULT', 'rangeWindowWidth') * 0.5
    maxRange = config.getfloat('DEFAULT', 'maxRange')  # for the plotting

    sphereTS = config.getfloat('DEFAULT', 'sphereTS')

    setupLogging(logDir, 'estimate_gains')

    sr = sonarReader(config)

    calLog = pd.read_csv(calLogFile, parse_dates=['start_time', 'end_time'],
                         dtype={'beam_name': str}, comment='#')

    calLog = calLog.assign(ts_mean=pd.Series(np.empty(calLog.shape[0])).values)
    calLog = calLog.assign(ts_rms=pd.Series(np.empty(calLog.shape[0])).values)
    calLog = calLog.assign(ts_range=pd.Series(np.empty(calLog.shape[0])).values)
    calLog = calLog.assign(ts_num=pd.Series(np.zeros(calLog.shape[0]), dtype=np.intc).values)
    calLog = calLog.assign(gain_new=pd.Series(np.empty(calLog.shape[0])).values)
    calLog = calLog.assign(gain_old=pd.Series(np.empty(calLog.shape[0])).values)

    for i, row in calLog.iterrows():
        if i >= 0:  # for testing. Lets us select particular rows
            r, ts, gainOld = sr.get_beam_TS(row['beam_name'], row['start_time'], row['end_time'])
            search_r = row['range']

            if r is None:
                continue

            ts_mean, ts_rms, ts_range, ts_num = sr.estimate_TS_at_range((search_r-offset,
                                                                         search_r+offset), r, ts)

            # and the new gain correction is....
            gainNew = gainOld + ts_mean - sphereTS

            # This is quite specific to the actual sonar equations - is ok for Simrad and Furuno
            # omnisonars as of 2024...
            ts_mean = ts_mean + gainOld - gainNew

            calLog.iloc[i, calLog.columns.get_loc('ts_mean')] = ts_mean
            calLog.iloc[i, calLog.columns.get_loc('ts_rms')] = ts_rms
            calLog.iloc[i, calLog.columns.get_loc('ts_range')] = ts_range
            calLog.iloc[i, calLog.columns.get_loc('ts_num')] = ts_num
            calLog.iloc[i, calLog.columns.get_loc('gain_old')] = gainOld
            calLog.iloc[i, calLog.columns.get_loc('gain_new')] = gainNew

            logging.info('  Beam %s has TS = %.1f with RMS of %.2f dB at %.1f m',
                         row["beam_name"], ts_mean, ts_rms, ts_range)

            fig, _ = plt.subplots()
            plt.plot(r, ts, linewidth=0.5)
            plt.plot([search_r-offset, search_r-offset], [-140, -20], 'k')
            plt.plot([search_r+offset, search_r+offset], [-140, -20], 'k')
            plt.text(0.75*maxRange, -23, f'TS = {ts_mean:.1f} dB')
            t = row['start_time'].strftime('%H:%M:%S')
            plt.title(f'Beam {row["beam_name"]} starting at {t}')
            plt.xlim(0, maxRange)
            plt.grid()
            plt.xlabel('Range [m]')
            plt.ylabel('TS [dB re 1 $m^2$]')
            fig.tight_layout(pad=3.0)

            # Save figure to an image
            t = row['start_time'].strftime('%H%M%S')
            fig.savefig(resultsDir.joinpath(f'Beam_{row["beam_name"]}_{t}.png'))

            # and close it (otherwise there are too many open at once)
            plt.close()

    calLog.to_csv(resultsDir.joinpath('results.csv'), index=False,
                  float_format='%.2f', date_format='%Y-%m-%dT%H:%M:%S')


class sonarReader:
    """Read omnisonar data files that are in the sonar-netCDF4 format."""

    def __init__(self, config):

        # Pull out the settings in the config file.
        self.beamGroup = config.get('DEFAULT', 'beamGroupPath')
        self.dataDir = Path(config.get('DEFAULT', 'dataDir'))

        # get ping times for all pings in all files in dataDir to make it
        # convenient to work out which files to use for each data request
        files = self.dataDir.glob('*.nc')

        self.file_start_times = []
        self.file_end_times = []
        self.filenames = []

        for ff in files:
            with h5py.File(ff, 'r') as f:
                start_time = f[self.beamGroup + '/ping_time'][0]
                end_time = f[self.beamGroup + '/ping_time'][-1]
                start_time = cftime.num2pydate(start_time/1e6,
                                               'milliseconds since 1601-01-01 00:00:00')
                end_time = cftime.num2pydate(end_time/1e6, 'milliseconds since 1601-01-01 00:00:00')
                self.file_start_times.append(start_time)
                self.file_end_times.append(end_time)
                self.filenames.append(ff)

        self.file_start_times = np.array(self.file_start_times)
        self.file_end_times = np.array(self.file_end_times)
        self.filenames = np.array(self.filenames)

    def estimate_TS_at_range(self, range_bounds, r, ts):
        """Calculate the mean TS from the strongest echoes in range_bound over all given pings."""
        logging.info('  Searching for maximum response between %.1f and %.1f m',
                     range_bounds[0], range_bounds[1])
        mask = (r >= range_bounds[0]) & (r <= range_bounds[1])
        ts_max = np.max(ts[mask, :], axis=0)

        # and we search again to get the index. Yes, this mean two searches when
        # it could be one search, but it will make things looks complicated to
        # do it all with argmax()
        ts_max_i = np.argmax(ts[mask, :], axis=0)
        ranges = np.array([r[i]+range_bounds[0] for i in ts_max_i])

        trimmed = ms.trim(ts_max, limits=(0.05, 0.05), relative=True)
        mask = np.logical_not(np.ma.getmaskarray(trimmed))
        ts_max_trimmed = ts_max[mask]

        ts_mean = 10.0 * np.log10(np.mean(np.power(10, ts_max_trimmed/10.0)))
        ts_rms = np.sqrt(np.mean(np.square(ts_max_trimmed-ts_mean)))
        ts_range = np.mean(ranges[mask])
        ts_num = len(ts_max_trimmed)

        return ts_mean, ts_rms, ts_range, ts_num

    def get_beam_TS(self, beamName, startTime, endTime):
        """Load the requested beam between the requested times."""
        logging.info('Processing beam "%s" from %s to %s', beamName, startTime, endTime)

        # Which files contain data between the start and end times?
        first_file_i = np.nonzero((startTime >= self.file_start_times)
                                  & (startTime <= self.file_end_times))[0]
        last_file_i = np.where((endTime >= self.file_start_times)
                               & (endTime <= self.file_end_times))[0]

        if len(first_file_i) == 0:
            logging.error('  Calibration start time is not covered by available data files')

        if len(last_file_i) == 0:
            logging.error('  Calibration end time is not covered by available data files')

        files = self.filenames[int(first_file_i[0]):int(last_file_i[0])+1]

        logging.info('  Using files %s', ', '.join([str(x.name) for x in files]))

        ts = []
        for file in files:
            f = h5py.File(file, 'r')
            logging.info('  Reading data from file %s', file.name)

            t = f[self.beamGroup + '/ping_time']
            tt = cftime.num2pydate(t[:]/1e6, 'milliseconds since 1601-01-01 00:00:00')
            beamIds = f[self.beamGroup + '/beam']
            beamIds = (b.decode() for b in beamIds)
            # Find index of beam named beamName
            beam = next((i for (i, name) in enumerate(beamIds) if name == beamName), None)
            if beam is None:
                logging.error(' No beam found with name: "%s"', beamName)
                return (None, None, None)

            # find the pings within the time period
            within = np.nonzero((tt >= startTime) & (tt <= endTime))[0]

            # Calculate and store the TS for each ping
            for ping_i in within:
                # convert x,y,z direction into a horizontal angle for use elsewhere
                x = f[self.beamGroup + '/beam_direction_x'][ping_i]
                y = f[self.beamGroup + '/beam_direction_y'][ping_i]
                z = f[self.beamGroup + '/beam_direction_z'][ping_i]
                tilt = np.arctan(z / np.sqrt(x**2 + y**2))  # [rad]
                r, ping, gain = self.TSFromSonarNetCDF4(f, self.beamGroup, ping_i, beam, tilt)

                if len(ts) == 0:
                    ts = ping
                else:
                    ts = np.vstack((ts, ping))

            f.close()

        return r, np.transpose(ts), gain  # uses the last r and gain's

    def TSFromSonarNetCDF4(self, f, beamGroup, ping, beam, tilt):
        """Calculate TS from the given beam group, beam, and ping."""
        eqn_type = f[beamGroup].attrs['conversion_equation_type']
        # work around the current Simrad files using integers instead of the
        # type defined in the convetion (which shows up here as a string)
        if isinstance(eqn_type, np.ndarray):
            eqn_type = f'type_{eqn_type[0]}'
        else:
            eqn_type = eqn_type.decode('utf-8')

        gain = 0  # Return the existing gain via this

        if eqn_type == 'type_2':
            # Pick out various variables for the given ping and beam
            ts = f[beamGroup + '/backscatter_r'][ping][beam]  # an array for each beam/ping

            SL = f[beamGroup + '/transmit_source_level'][ping]  # a scalar for the current ping
            K = f[beamGroup + '/receiver_sensitivity'][ping][beam]  # a scalar for each beam/ping
            deltaG = f[beamGroup + '/gain_correction'][ping][beam]  # a scalar for each beam
            G_T = f[beamGroup + '/time_varied_gain'][ping]  # per sample in the current ping

            ping_freq_1 = f[beamGroup + '/transmit_frequency_start'][ping][beam]  # scalar per beam
            ping_freq_2 = f[beamGroup + '/transmit_frequency_stop'][ping][beam]  # scalar per beam

            # and some more constant things that could be moved out of this function...
            c = f['Environment/sound_speed_indicative'][()]  # a scalar
            alpha_vector = f['Environment/absorption_indicative'][()]  # a vector
            freq_vector = f['Environment/frequency'][()]  # a vector
            ping_freq = (ping_freq_1 + ping_freq_2)/2.0  # a scalar for each beam
            alpha = np.interp(ping_freq, freq_vector, alpha_vector)  # a scalar for each beam

            # some files have nan for some of the above variables, so fix that
            if np.any(np.isnan(deltaG)):
                deltaG = np.zeros(deltaG.shape)
            if np.any(np.isnan(alpha_vector)):
                # quick and dirty...
                alpha = self.acousticAbsorption(10.0, 35.0, 10.0, ping_freq)

            a = SL + K + deltaG + G_T  # a scalar
            samInt = f[beamGroup + '/sample_interval'][ping]  # [s]

            gain = deltaG

            # Ignore some errors as there are usually some zeros in the data of no real consequence
            with np.errstate(divide='ignore', invalid='ignore'):
                # [m] range vector for the current beam/ping
                r = samInt * c/2.0 * np.arange(0, ts.size)
                ts = 20.0*np.log10(ts/np.sqrt(2.0)) + 40.0*np.log10(r) + 2*alpha*r - a
        elif eqn_type == 'type_1':
            # Pick out various variables for the given ping, i
            p_r = f[beamGroup + '/backscatter_r'][ping][beam]  # an array for each beam/ping
            p_i = f[beamGroup + '/backscatter_i'][ping][beam]  # an array for each beam/ping
            ts = np.absolute(p_r + 1j*p_i)
            G = f[beamGroup + '/transducer_gain'][ping][beam]  # a scalar for each beam
            P = f[beamGroup + '/transmit_power'][ping]  # a scalar
            ping_freq_1 = f[beamGroup + '/transmit_frequency_start'][ping]  # a scalar for each beam
            ping_freq_2 = f[beamGroup + '/transmit_frequency_stop'][ping]  # a scalar for each beam

            # and some more constant things that could be moved out of this function...
            c = f['Environment/sound_speed_indicative'][()]  # a scalar
            alpha_vector = f['Environment/absorption_indicative'][()]  # a vector
            freq_vector = f['Environment/frequency'][()]  # a vector
            ping_freq = (ping_freq_1 + ping_freq_2)/2.0  # a scalar for each beam
            alpha = np.interp(ping_freq, freq_vector, alpha_vector)  # a scalar
            wl = c / ping_freq  # wavelength [m]

            if np.any(np.isnan(alpha_vector)):
                # quick and dirty...
                alpha = self.acousticAbsorption(10.0, 35.0, 10.0, ping_freq)

            samInt = f[beamGroup + '/sample_interval'][ping]  # [s]

            r_offset = 0.0  # incase we need this in the future

            gain = G

            # Ignore some errors as there are usually some zeros in the data of no real consequence
            with np.errstate(divide='ignore', invalid='ignore'):
                # [m] range vector for the current beam
                r = samInt * c/2.0 * np.arange(0, ts.size) - r_offset
                ts = 20.0*np.log10(ts) + 40.0*np.log10(r) + 2*alpha*r\
                    - 10.0*np.log10((P*wl*wl) / (16*np.pi*np.pi))\
                    - G - 40.0*np.log10(np.cos(tilt[beam]))

        else:  # unsupported format - just take the log10 of the numbers. Usually usefull.
            ts = f[beamGroup + '/backscatter_r'][ping]
            with np.errstate(divide='ignore'):
                for j in range(0, ts.shape[0]):
                    ts[j] = np.log10(ts[j])
                    r = np.array([])  # Fix this

        return r, ts, gain

    def acousticAbsorption(self, temperature, salinity, depth, frequency):
        """Calculate acoustic absorption.

        Uses Ainslie & McColm, 1998.
        Units are:
            temperature - degC
            salinity - PSU
            depth - m
            frequency - Hz
            alpha - dB/m
        """
        frequency = frequency / 1e3  # [kHz]
        pH = 8.0

        z = depth/1e3  # [km]
        f1 = 0.78 * np.sqrt(salinity/35.0) * np.exp(temperature/26.0)
        f2 = 42.0 * np.exp(temperature/17.0)
        alpha = 0.106 * (f1*frequency**2./(frequency**2+f1**2)) * np.exp((pH-8.0)/0.56) \
            + 0.52*(1+temperature/43.0) * (salinity/35.0) \
            * (f2*frequency**2)/(frequency**2+f2**2) * np.exp(z/6.0) \
            + 0.00049*frequency**2 * np.exp(-(temperature/27.0+z/17.0))
        alpha = alpha * 1e-3  # [dB/m]

        return alpha


def setupLogging(log_dir, label):
    """Set info, warning, and error message logger to a file and to the console."""
    now = datetime.now(timezone.utc)
    logger_filename = os.path.join(log_dir, now.strftime('log_' + label + '-%Y%m%d-T%H%M%S.log'))
    logger = logging.getLogger('')

    # Add handlers if none are present
    if not logging.getLogger('').hasHandlers():
        logger.setLevel(logging.INFO)

        formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')

        # A logger to a file that is changed periodically
        rotatingFile = logging.handlers.TimedRotatingFileHandler(logger_filename, when='H',
                                                                 interval=12, utc=True)
        rotatingFile.setFormatter(formatter)
        logger.addHandler(rotatingFile)

        # add a second output to the logger to direct messages to the console
        console = logging.StreamHandler()
        console.setFormatter(formatter)
        logger.addHandler(console)

        logging.info('Log files are in %s', log_dir.as_posix())


if __name__ == "__main__":
    main()



import pandas as pd

def calculate_gain(sphere_ts):
    """Calculate the beam calibration gain and other stats."""
    df = pd.DataFrame(sphere_ts, columns=['timestamp', 'ts', 'range'])
    return (df['ts'].mean(), df['ts'].std(), df['range'].mean(), len(df))





import webbrowser
import logging
from pathlib import Path
from datetime import datetime
import tkinter as tk
from tkinter import ttk
from tkinter import filedialog as fd
from tkinter import messagebox
from importlib.metadata import version
from PIL import Image, ImageTk
from .utils import window_closed, app_name
from .calibration_data import calibrationData
from .calculate_gains import calculate_gain
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from platformdirs import PlatformDirs

logger = logging.getLogger(app_name)
icon_file = Path(__file__).parent.parent.parent/'assets'/'icon.png'  # TODO get via a config file

class calibrationGUI:
    """Provides the main GUI container and misc labels/buttons."""

    def __init__(self, echogram, title='TITLE', help_uri=None):
        self.echogram = echogram
        self.help_uri = help_uri

        # Calibration gains are stored in here
        self.cal_data = calibrationData()
        # sphere ts for the current beam calibration is stored in here
        self.sphere_ts = []

        # The GUI window
        self.echogram.root.title(title)
        
        # Dialogs that we keep around
        self.results_dialog = None
        
        # The toolbar and window icon/logo
        self.icon = ImageTk.PhotoImage(Image.open(icon_file))
        self.echogram.root.iconphoto(False, self.icon)

        # Things to do with new pings 
        self.echogram.set_ping_callback(self.new_ping)

        # Put the matplotlib plots into the GUI window.
        canvas = FigureCanvasTkAgg(self.echogram.fig, master=self.echogram.root)
        canvas.get_tk_widget().pack(side='top', fill='both', expand=True, padx=5, pady=5)

        # Styles. These apply to all widgets, not just the ones create in this function
        s = ttk.Style()
        s.configure('TButton', font=('Arial', 16))
        s.configure('TLabel', font=('Arial', 12))
        s.configure('TCheckbutton', font=('Arial', 16))
        s.configure('Treeview.Heading', font=('Arial', 12, 'bold'))
        s.configure('Treeview', font=('Arial', 12))

        # A label to show the last received message time
        self.label = ttk.Label(self.echogram.root)
        self.label.pack(side=tk.TOP, fill=tk.BOTH)
        self.label.config(text='Waiting for data...', width=100, anchor=tk.W)

        ttk.Separator(self.echogram.root, orient='horizontal').pack(fill='x', padx=10, pady=5)

        # Buttons for help, on-axis toggle, config dialog, and close
        self.onaxis_value = tk.BooleanVar(value=False)

        frame = ttk.Frame(self.echogram.root)
        results = ttk.Button(frame, text='Results', command=self.results)
        config = ttk.Button(frame, text='Config', command=self.config)
        onaxis = ttk.Checkbutton(frame, text='On-axis', variable=self.onaxis_value,
                                 command=self.onaxis_changed)
        help = ttk.Button(frame, text='Help', command=self.help)
        about = ttk.Button(frame, text='About', command=self.about)
        close = ttk.Button(frame, text='Close', command=self.close)

        onaxis.pack(side=tk.LEFT)
        close.pack(side=tk.RIGHT)
        about.pack(side=tk.RIGHT)
        help.pack(side=tk.RIGHT)
        config.pack(side=tk.RIGHT)
        results.pack(side=tk.RIGHT)

        frame.pack(side=tk.TOP, fill=tk.BOTH)

        # Start listening for sonar data
        self.echogram.newPing(self.status_label())

    def job(self):
        return self.echogram.job

    def root(self):
        return self.echogram.root

    def onaxis_changed(self):
        """A beam calibration has either started or ended."""
        if self.onaxis_value.get():  # start calibrating a beam
            self.echogram.beamLine.freeze(True)
            logger.info('Beam %d calibration started', self.echogram.beam)
        else:  # finished calibrating a beam
            logger.info('Beam %d calibration complete', self.echogram.beam)
            self.echogram.beamLine.freeze(False)
            self.sphere_ts = []
            if self.results_dialog:
                self.results_dialog.update_rows(None)  # unhighlights the previously active row

    def new_ping(self):
        """Orchestrates things for each new ping."""
        e = self.echogram
        if e.beamLine.frozen():  # a beam is being calibrated
            # store the current ping's sphere echo info
            self.sphere_ts.append((datetime.now().isoformat(), e.amp[1, -1], e.rangeMax))
            # calculate the beam gain and other stats
            (gain, rms, r, num) = calculate_gain(self.sphere_ts)
            # store the latest beam gain values
            self.cal_data.update(e.beam, datetime.now().strftime('%H:%M:%S'), gain, rms, r, num)
            # update the results dialog if present
            if self.results_dialog:
                self.results_dialog.update_with(self.cal_data, e.beam)

    def about(self):
        message = (f'Sonarcal, version {version("sonarcal")}\n\n'
                   'A program to assist with calibrating omni-directional sonars.\n\n'
                   'Developed by Aqualyd Ltd\n\n'
                   'www.aqualyd.nz')

        messagebox.showinfo(title='About', message=message)

    def close(self):
        window_closed(self.echogram.root, self.echogram.job)

    def results(self):
        """Open the Results dialog box."""
        # want one lasting instance of this dialog so manage that here
        if not self.results_dialog:
            self.results_dialog = resultsDialog(self.echogram.root, self.cal_data, self.icon)
        else:
            self.results_dialog.reopen()

    def help(self):
        """Open the help documentation in a web browser."""
        if not webbrowser.open(self.help_uri, new=2):
            logging.warning('Failed to start a webbrowser to show the help documentation')
        
    def config(self):
        """Open the Config dialog box."""
        configDialog(self.echogram.root, self.icon)

    def status_label(self):
        return self.label


class configDialog:
    """A dialog box to set and change application parameters."""
    def __init__(self, parent, icon=None):
        self.top = tk.Toplevel(parent)
        self.top.title("Config")
        if icon:
            self.top.iconphoto(False, icon)

        ttk.Label(self.top, text="Configs").pack(padx=20, pady=10)
        ttk.Button(self.top, text="Close", command=self.close_dialog).pack(pady=5)
        ttk.Button(self.top, text="Apply", command=self.apply).pack(pady=5)

    def apply(self):
        pass

    def close_dialog(self):
        self.top.destroy()


class resultsDialog:
    """A dialog box to show completed calibration results per beam."""

    def __init__(self, parent, data: dict=None, icon=None):
        
        self.data = data
        
        self.top = tk.Toplevel(parent)
        self.top.title("Results")
        if icon:
            self.top.iconphoto(False, icon)

        tree_frame = ttk.Frame(self.top)

        # use a ttk.Treeview to show a table of the results
        self.item_ids = {}  # contains ids to rows that get added to the treeview
        self.setup_treeview(tree_frame, self.data)  # creates self.tree

        # Make scrollbars for the treeview widget
        vsb = ttk.Scrollbar(tree_frame, orient="vertical", command=self.tree.yview)
        self.tree.configure(yscrollcommand=vsb.set)
        
        # pack the scrollbars and treeview
        vsb.pack(side="right", fill="y")
        self.tree.pack(side=tk.TOP, fill=tk.BOTH, expand=True, padx=5)

        # create and pack the buttons
        btn_frame = ttk.Frame(self.top)
        remove = ttk.Button(btn_frame, text="Remove selected", command=self.remove_rows)
        save = ttk.Button(btn_frame, text="Save", command=self.save)
        close = ttk.Button(btn_frame, text="Close", command=self.close_dialog)
        
        close.pack(side=tk.RIGHT)
        save.pack(side=tk.RIGHT)
        remove.pack(side=tk.RIGHT)

        tree_frame.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        btn_frame.pack(side=tk.BOTTOM, fill=tk.X)

    def setup_treeview(self, top, data):
        """Create the treeview columns, etc."""
        
        headings = [data.df().index.name] + list(data.df().columns)
        
        self.tree = ttk.Treeview(top, columns=headings, show='headings')
        self.tree.bind('<Button-1>', self.on_row_click)

        # colour odd and even rows
        self.tree.tag_configure('evenrow', background='white smoke')
        self.tree.tag_configure('oddrow', background='white')
        self.tree.tag_configure('active', background='orange2')
        
        for col in headings:
            self.tree.heading(col, text=col, anchor='e')
            self.tree.column(col, width=100, anchor='e')

        # Add rows (if any)
        self.update_with(data)            
                    
    def update_with(self, data, active_beam: int|None = None):
        """Update dialog's display with given calibration data."""

        for beam, row in data.df().iterrows():

            # TODO
            # This is hacky and fragile - find a better way that doesn't need to name the
            # columns
            values = [beam, 
                      row['Time'],
                      f"{row['Gain (dB)']:0.1f}",
                      f"{row['RMS (dB)']:0.1f}",
                      f"{row['Range (m)']:0.1f}",
                      row['Echoes']]
            # values = [beam] + list(row)

            if beam in self.item_ids:
                # row for this beam already exists, so update it
                self.tree.item(self.item_ids[beam], values=values)
            else:
                # new beam, so add a row
                item_id = self.tree.insert('', 'end', values=values)
                self.item_ids[beam] = item_id

        self.update_rows(active_beam)

        # keep this to use in the save and remove functionalities
        self.data = data

    def update_rows(self, active_beam: int|None):
        """Sorts calibration results rows by beam and sets background colours to look nice."""
        for beam, index in zip(sorted(self.item_ids.keys()), range(len(self.item_ids))):
            self.tree.move(self.item_ids[beam], '', index)

            rowness = 'oddrow' if index % 2 == 0 else 'evenrow'
            if active_beam and beam == active_beam:
                rowness = 'active'

            self.tree.item(self.item_ids[beam], tags=(rowness,))    

    def remove_rows(self):
        """Remove selected rows from the results table."""

        selected = self.tree.selection()  # returns a tuple of item_ids
        
        # Pick out the beam numbers of the selected rows
        to_remove = []
        for iid in selected:
            item_data = self.tree.item(iid)
            values = item_data['values']
            to_remove.append(values[0])

        if to_remove:
            for beam in to_remove:
                # remove selected rows from treeview
                self.tree.delete(self.item_ids[beam])
                # remove beam imtem from the map between beam and item ids
                self.item_ids.pop(beam)
            # remove beams from the cal_data store
            self.data.remove(to_remove)
            self.update_rows(None)

    def save(self):
        """Save the results to a file."""
        timestamp = datetime.now().strftime('%Y%m%dT%H%M%S')
        default_filename = 'sonar_calibration_' + timestamp + '.csv'
        save_filename = fd.asksaveasfilename(title='Save as CSV', defaultextension='.csv',
                                             initialdir=PlatformDirs.user_documents_dir,
                                             initialfile=default_filename,
                                             filetypes=[('CSV', '*.csv')])
        if save_filename:
            logger.info('Saved results to %s', save_filename)
            self.data.df().to_csv(save_filename)

    def on_row_click(self, event):
        """Implement selection and deselection."""
        # TODO - work out why it can take multiple clicks on a row to get it unselected
        item_id = self.tree.identify_row(event.y)
        if item_id and item_id in self.tree.selection():
            self.tree.selection_remove(item_id)


    def reopen(self):
        self.top.deiconify()

    def close_dialog(self):
        self.top.withdraw()
